import signal
from collections import defaultdict
from typing import List
from pathlib import Path
import logging
import hashlib
import re
import time

import toml

_DEBUG = False
l = logging.getLogger(__name__)


class timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type_, value, traceback):
        signal.alarm(0)


def _get_angr_dec_options(param_string):
    import angr
    opt_candidates = [
        o for o in angr.analyses.decompiler.decompilation_options.options
        if o.param == param_string
    ]

    if not opt_candidates:
        return None

    return opt_candidates[0]


def angr_comb_decompile(binary_path, functions=None, print_dec=False, **kwargs):
    return angr_decompile(binary_path, functions=functions, print_dec=print_dec, structurer="combing", use_sailr=False, name="angr_comb", improved_restruct=False, **kwargs)


def angr_improved_phoenix_decompile(binary_path, functions=None, print_dec=False, **kwargs):
    return angr_decompile(binary_path, functions=functions, print_dec=print_dec, structurer="phoenix", use_sailr=False, name="angr_improved_phoenix", improved_restruct=True, **kwargs)


def angr_phoenix_decompile(binary_path, functions=None, print_dec=False, **kwargs):
    return angr_decompile(binary_path, functions=functions, print_dec=print_dec, structurer="phoenix", use_sailr=False, name="angr_phoenix", improved_restruct=False, **kwargs)


def angr_dream_decompile(binary_path, functions=None, print_dec=False, **kwargs):
    return angr_decompile(binary_path, functions=functions, print_dec=print_dec, structurer="dream", use_sailr=False, name="angr_dream", improved_restruct=False, **kwargs)


def angr_decompile(
        binary_path,
        functions=None,
        print_dec=False,
        structurer="phoenix",
        use_sailr=True,
        func_dec_timeout=600,
        name="angr_sailr",
        improved_restruct=True,
        dump_line_maps=True,
        dump_early_metrics=True,
        symboless=False,
        disable_opts=[]
):
    #
    # Do dynamic imports
    #

    import angr
    from angr.analyses.decompiler.optimization_passes import EagerReturnsSimplifier, LoweredSwitchSimplifier, \
        ReturnDeduplicator
    from cle.backends.coff import Coff
    try:
        from angr.analyses.decompiler.optimization_passes import (
            DuplicationOptReverter, ConstPropOptReverter, CrossJumpReverter, FlipBooleanCmp, ITERegionConverter,
            ReturnDeduplicator
        )
    except ImportError:
        DuplicationOptReverter = 1
        ConstPropOptReverter = 2
        CrossJumpReverter = 3
        FlipBooleanCmp = 4
        ITERegionConverter = 5
        ReturnDeduplicator = 6

    #
    # real code
    #

    angr_name = name
    proj = angr.Project(binary_path, auto_load_libs=False)
    is_windows = proj.loader.main_object.os == "windows" or isinstance(proj.loader.main_object, Coff)
    cfg = proj.analyses.CFG(show_progressbar=False, normalize=True, data_references=True)
    try:
        proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=True, analyze_callsites=True)
        cc_failed = False
    except Exception:
        cc_failed = True

    if cc_failed:
        l.warning(f"CallingConvention Analysis failed on {binary_path}. Trying again without variable recovery...")
        try:
            # try it again without variable recovery
            proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=False)
            cc_failed = False
        except Exception:
            pass

    if cc_failed:
        l.critical(f"All attempts to run CallingConvention Analysis failed on {binary_path}.")

    all_optimization_passes = angr.analyses.decompiler.optimization_passes.get_default_optimization_passes("AMD64", "linux")

    # decide to use options in SAILR package
    opt_blacklist = [] if use_sailr else [
        DuplicationOptReverter, ConstPropOptReverter, CrossJumpReverter, EagerReturnsSimplifier,
        LoweredSwitchSimplifier, FlipBooleanCmp, ITERegionConverter, ReturnDeduplicator
    ]
    if is_windows and LoweredSwitchSimplifier not in opt_blacklist:
         opt_blacklist.append(LoweredSwitchSimplifier)

    if disable_opts:
        loaded_opts = [
            globals()[opt] for opt in disable_opts if opt in globals()
        ]
        for opt in loaded_opts:
            if opt not in opt_blacklist:
                opt_blacklist.append(opt) 

    if improved_restruct:
        # improves structuring gets boolean flipper
        if FlipBooleanCmp in opt_blacklist:
            opt_blacklist.remove(FlipBooleanCmp)

    all_optimization_passes = [
        p for p in all_optimization_passes
        if p not in opt_blacklist
    ]

    binary_path = Path(binary_path).absolute()
    bin_name = binary_path.with_suffix("").name
    output_path = binary_path.parent.joinpath(f"{angr_name}_{bin_name}.c").absolute()
    functions = functions or cfg.functions
    decompilation = ""

    # clean up function names
    for func in list(cfg.functions.values()):
        if "." in func.name:
            func.name = func.name[:func.name.index(".")]

    # decide to use options
    options = [
        (_get_angr_dec_options("structurer_cls"), structurer),
        (_get_angr_dec_options("comment_gotos"), False),
        (_get_angr_dec_options("braces_on_own_lines"), False),
        (_get_angr_dec_options("show_local_types"), False),
        (_get_angr_dec_options("show_externs"), False),
        (_get_angr_dec_options("show_casts"), False),
        (_get_angr_dec_options("improve_structurer"), improved_restruct),
        (_get_angr_dec_options("largest_successor_tree_outside_loop"), improved_restruct),
        (_get_angr_dec_options("simplify_switches"), use_sailr),
    ]
    # add extra option which may not be available based on commit version
    simplify_ifelse_opt = _get_angr_dec_options("simplify_ifelse")
    if simplify_ifelse_opt is not None:
        options.append((simplify_ifelse_opt, improved_restruct))

    linemaps = {}
    metrics = {}
    for func_idx, func_addr in enumerate(functions):
        try:
            f = cfg.functions[func_addr]
        except Exception:
            continue

        if f is None:
            continue

        if f.is_plt:
            continue

        starttime = time.time()
        exception_string = ""
        if _DEBUG:
            dec = proj.analyses.Decompiler(f, cfg=cfg, kb=cfg.kb, options=options, optimization_passes=all_optimization_passes)
        else:
            try:
                l.info(f"Decompiling {bin_name}.{f.name} now ({func_idx}/{len(functions)})...")
                with timeout(seconds=func_dec_timeout):
                    dec = proj.analyses.Decompiler(f, cfg=cfg, kb=cfg.kb, options=options, optimization_passes=all_optimization_passes)
            except Exception as e:
                l.critical(f"Failed to decompile {bin_name}.{f.name} because {e}")
                exception_string = str(e).replace("\n", " ")
                dec = None

        # exception forcefully crashed
        has_error = exception_string != ""

        #
        # do sanity checks on decompilation, skip checks if we already errored
        #

        # codegen failed in decompilation but did not crash, probably in structuring
        if not has_error and (dec is None or not dec.codegen or not dec.codegen.text):
            exception_string = "Decompilation had no code output (failed in Dec)"
            has_error = True

        # codegen created code, but it was empty (likely structuring error)
        if not has_error and ("{\n}" in dec.codegen.text):
            exception_string = "Decompilation outputted an empty function (failed in structuring)"
            has_error = True

        # codegen created corrupted code
        #if not has_error and ("[D]" in dec.codegen.text or "DIRTY" in dec.codegen.text):
        #    exception_string = "Decompilation outputted a dirty function (failed in lifting)"
        #    has_error = True

        if not has_error and (structurer in ["dream", "combing"] and "goto" in dec.codegen.text):
            exception_string = "Decompilation outputted a goto for Gotoless algorithm (failed in structuring)"
            has_error = True

        # skip errored decompilation
        if has_error:
            l.critical(f"Failed to decompile {bin_name}.{f.name} because {exception_string}")
            decompilation += f"// [error: {f.name} | {exception_string}]\n"
            continue

        # decompilation must be valid
        delta_time = time.time() - starttime
        decompilation += dec.codegen.text + "\n"
        func_name = f.name if not symboless else hex(f.addr)

        # dump metrics
        if dump_line_maps and not print_dec:
            linemaps[func_name] = generate_linemaps(dec, dec.codegen, base_addr=proj.loader.main_object.min_addr)
        if dump_early_metrics and not print_dec:
            metrics[func_name] = collect_countable_metrics(dec.codegen)
            metrics[func_name]["dec_time"] = delta_time

    if print_dec:
        print(decompilation)
        return decompilation

    with open(output_path, "w") as fp:
        fp.write(decompilation)

    #
    # metric and extras dumping
    #

    if dump_line_maps:
        final_linemaps = {k: dict(v) for k,v in linemaps.items()}
        dump_path = output_path.with_suffix(".linemaps")
        with open(dump_path, "w") as fp:
            toml.dump(final_linemaps, fp)
    if dump_early_metrics:
        dump_path = output_path.with_suffix(".toml")
        with open(dump_path, "w") as fp:
            toml.dump(metrics, fp)

    return output_path


def generate_linemaps(dec, codegen, base_addr=0x400000):
    import ailment
    from angr.analyses.decompiler.structured_codegen.c import CStructuredCodeWalker, CFunctionCall, CIfElse, CIfBreak

    if codegen is None:
        return
    
    base_addr = dec.project.loader.main_object.image_base_delta
    if hasattr(dec, "unmodified_clinic_graph"):
        nodes = dec.unmodified_clinic_graph.nodes
    else:
        l.warning(f"You are likely using an older version of angr that has no unmodified_clinic_graph."
                  f" Using clinic_graph instead, results will be less accurate...")
        nodes = dec.clinic.cc_graph.nodes

    # get the mapping of the original AIL graph
    mapping = defaultdict(set)
    ail_node_addr_map = {
        node.addr: node for node in nodes
    }
    for addr, ail_block in ail_node_addr_map.items():
        # get instructions of this block
        try:
            vex_block = dec.project.factory.block(addr)
        except Exception:
            continue

        ail_block_stmts = [stmt for stmt in ail_block.statements if not isinstance(stmt, ailment.statement.Label)]
        if not ail_block_stmts:
            continue

        next_ail_stmt_idx = 0
        for ins_addr in vex_block.instruction_addrs:
            next_ail_stmt_addr = ail_block_stmts[next_ail_stmt_idx].ins_addr
            mapping[next_ail_stmt_addr].add(ins_addr)
            if ins_addr == next_ail_stmt_addr:
                next_ail_stmt_idx += 1
            if next_ail_stmt_idx >= len(ail_block_stmts):
                break

    # node to addr map
    ailaddr_to_addr = defaultdict(set)
    for k, v in mapping.items():
        for v_ in v:
            ailaddr_to_addr[k - base_addr].add(v_ - base_addr)

    """
    ail_node_addr_map = {
        node.addr: node for node in dec.clinic.cc_graph
    }
    """

    codegen.show_externs = False
    codegen.regenerate_text()

    decompilation = codegen.text
    if not decompilation:
        return
    
    try:
        first_code_pos = codegen.map_pos_to_addr.items()[0][0]
    except Exception:
        return

    # map the position start to an address
    pos_addr_map = defaultdict(set)
    for start, pos_map in codegen.map_pos_to_addr.items():
        obj = pos_map.obj
        if not hasattr(obj, "tags"):
            continue

        # leads to mapping at the beginning of loops, so skip.
        # see kill.o binary for send_signals
        if isinstance(obj, CIfElse):
            continue

        ins_addr = obj.tags.get("ins_addr", None)
        if ins_addr:
            pos_addr_map[start].add(ins_addr - base_addr)

        """
        vex_block_addr = obj.tags.get("vex_block_addr", None)
        if vex_block_addr == 0x400b59:
            breakpoint()

        if ins_addr is None:
            pos_addr_map[start].add(vex_block_addr - base_addr)
            continue

        ail_block = ail_node_addr_map.get(vex_block_addr, None)
        if ail_block is None:
            continue

        # get instructions of this block
        vex_block = dec.project.factory.block(vex_block_addr)
        ail_block_stmts = [stmt for stmt in ail_block.statements if not isinstance(stmt, ailment.statement.Label)]
        if not ail_block_stmts:
            continue

        next_ail_stmt_idx = 0
        for ins_addr in vex_block.instruction_addrs:
            next_ail_stmt_addr = ail_block_stmts[next_ail_stmt_idx].ins_addr
            mapping[next_ail_stmt_addr].add(ins_addr)
            if ins_addr == next_ail_stmt_addr:
                next_ail_stmt_idx += 1
            if next_ail_stmt_idx >= len(ail_block_stmts):
                break
        """


    # find every line
    line_end_pos = [i for i, x in enumerate(decompilation) if x == "\n"]
    line_to_addr = defaultdict(set)
    last_pos = len(decompilation) - 1
    line_to_addr[str(1)].add(codegen.cfunc.addr - base_addr)
    for i, pos in enumerate(line_end_pos[:-1]):
        if pos == last_pos:
            break

        curr_end = line_end_pos[i+1] - 1
        # check if this is the variable decs and header
        if curr_end < first_code_pos:
            line_to_addr[str(i+2)].add(codegen.cfunc.addr - base_addr)
            continue

        # not header, real code
        for p_idx in range(pos+1, curr_end+1):
            if p_idx in pos_addr_map:
                # line_to_addr[str(i+1)].update(pos_addr_map[p_idx])
                for ail_ins_addr in pos_addr_map[p_idx]:
                    if ail_ins_addr in ailaddr_to_addr:
                        line_to_addr[str(i+2)].update(ailaddr_to_addr[ail_ins_addr])
                    else:
                        line_to_addr[str(i+2)].add(ail_ins_addr)

    return line_to_addr


def count_cmove_insns(func):
    """
    func must be an angr function. Counts the number of cmove instructions in the function.
    """
    cmove_cnt = 0
    for block in func.blocks:
        for insn in block.disassembly.insns:
            if "cmove" in insn.mnemonic:
                cmove_cnt += 1

    return cmove_cnt


def collect_countable_metrics(codegen):
    from angr.analyses.decompiler.structured_codegen.c import CStructuredCodeWalker, CFunctionCall, CIfElse, CIfBreak
    #
    # gotos & bools
    #
    func_text = codegen.text
    gotos = re.findall("goto [a-zA-Z0-9_]*;", func_text, re.DOTALL)
    goto_cnt = len(gotos)
    and_cnt = func_text.count("&&")
    or_cnt = func_text.count("||")

    counts = {
        "loc": len(func_text.split("\n")),
        "gotos": goto_cnt,
        "bools": {
            "ands": and_cnt,
            "ors": or_cnt
        }
    }

    #
    # function calls
    #
    func_call_counts = defaultdict(int)
    class FunctionCallCounter(CStructuredCodeWalker):
        @classmethod
        def handle_CFunctionCall(cls, obj: CFunctionCall):
            if obj and obj.callee_func is not None:
                func_call_counts[obj.callee_func.name] += 1
            return super().handle_CFunctionCall(obj)

    FunctionCallCounter.handle(codegen.cfunc)
    func_call_counts = dict(func_call_counts)
    if not func_call_counts:
        func_call_counts = {"__empty__": 0}
    counts['func_calls'] = func_call_counts

    return counts


def angr_disassemble(binary_path, functions=None):
    import angr

    proj = angr.Project(binary_path, auto_load_libs=False)
    cfg = proj.analyses.CFG(show_progressbar=False, normalize=True)
    proj.analyses.CompleteCallingConventions(
        recover_variables=True,
        cfg=cfg,
        analyze_callsites=True,
        kb=cfg.kb,
    )

    functions = functions or cfg.functions
    return {
        func_name: cfg.functions[func_name].graph for func_name in functions
    }


def get_all_funcs(binary_path):
    import angr

    proj = angr.Project(binary_path, auto_load_libs=False)
    if not hasattr(proj.loader.main_object, "symbols_by_name"):
        return []

    syms: List[str] = [
        name for name, sym in proj.loader.main_object.symbols_by_name.items()
        if sym.is_function and not sym.is_import
    ]
    #cfg = proj.analyses.CFG(show_progressbar=False, normalize=True)
    #proj.analyses.CompleteCallingConventions(
    #    recover_variables=True,
    #    cfg=cfg,
    #    analyze_callsites=True,
    #    kb=cfg.kb,
    #)

    final_syms = []
    for sym in syms:
        if ".const" in sym:
            sym = sym.split(".const")[0]

        final_syms.append(sym)

    return final_syms


def _insn_bytes_from_func(function):
    insn_bytes = bytes()

    for blk_addr in sorted(function.block_addrs):
        blk = function.get_block(blk_addr)
        insn_bytes += blk.bytes

    return insn_bytes


def hash_functions(binary_path):
    import angr

    function_hash_dic = {}

    proj = angr.Project(binary_path, auto_load_libs=False)
    cfg = proj.analyses.CFGFast()

    functions = [func for addr, func in cfg.kb.functions.items()]

    for function in functions:
        function_insn_bytes = _insn_bytes_from_func(function)
        sha512_hash = hashlib.sha512(function_insn_bytes).hexdigest()
        function_hash_dic[function.name] = sha512_hash

    return function_hash_dic



from typing import List
import itertools
import sys

from .graph_utils import get_graph_root_nodes

import networkx as nx

memory = {}
ctr = itertools.count()


class DummyNode:
    def __init__(self):
        self.addr = next(ctr)
        self.path_count = 0

    def __eq__(self, other):
        return isinstance(other, DummyNode) and self.addr == other.addr

    def __hash__(self):
        return hash(self.addr)




#def count_paths(g, start_node, end_node):
#    if end_node == start_node:
#        return 1
#
#    if end_node in memory:
#        return memory[end_node]
#
#    nodes = g.predecessors(end_node)
#    total = 0
#    for node in nodes:
#        r = count_paths(g, start_node, node)
#        total += r
#    if isinstance(end_node, DummyNode):
#        total *= end_node.path_count
#    memory[end_node] = total
#    return total


#def count_total_paths(graph: nx.DiGraph):
#    global memory, ctr
#    memory = {}
#    ctr = itertools.count()
#
#    for component in list(nx.strongly_connected_components(graph)):
#        if len(component) > 1:
#            # create a subgraph
#            subgraph = graph.subgraph(component).copy()
#
#            all_predecessors = set()
#            all_successors = set()
#            loop_heads = []
#            for node in component:
#                for pred in graph.predecessors(node):
#                    if pred not in component:
#                        all_predecessors.add(pred)
#                        loop_heads.append(node)
#                for succ in graph.successors(node):
#                    if succ not in component:
#                        all_successors.add(succ)
#            new_node = DummyNode()
#            for pred in all_predecessors:
#                graph.add_edge(pred, new_node)
#            for succ in all_successors:
#                graph.add_edge(new_node, succ)
#            for nn in component:
#                graph.remove_node(nn)
#
#            # count the number of paths in the subgraph
#            # break the loop first
#            loop_head = loop_heads[0]
#            pred = list(subgraph.predecessors(loop_head))[0]
#            subgraph.remove_edge(pred, loop_head)
#            path_count = count_paths(subgraph, loop_head, pred)
#
#            # each loop iterates at most 6 times due to __nb_steps being 6 in this example program
#            path_count = path_count ** 6
#            new_node.path_count = path_count
#
#    # count number of paths
#    start_nodes = [node for node in graph.nodes if graph.in_degree(node) == 0]
#    start_nodes = get_graph_root_nodes(graph) if len(start_nodes) > 1 else start_nodes
#    start_node = start_nodes[0]
#    end_nodes = [node for node in graph.nodes if graph.out_degree(node) == 0]
#
#    total_paths = 0
#    for end_node in end_nodes:
#        total_paths += count_paths(graph, start_node, end_node)
#
#    return total_paths


def _all_simple_paths_graph(G, source, targets, cutoff):
    size = 0
    visited = dict.fromkeys([source])
    stack = [iter(G[source])]
    while stack:
        children = stack[-1]
        child = next(children, None)
        if child is None:
            stack.pop()
            visited.popitem()
        elif len(visited) < cutoff:
            if child in visited:
                continue
            if child in targets:
                size += len(list(visited)) + 1
            visited[child] = None
            if targets - set(visited.keys()):  # expand stack until find all targets
                stack.append(iter(G[child]))
            else:
                visited.popitem()  # maybe other ways to child
        else:  # len(visited) == cutoff:
            for target in (targets & (set(children) | {child})) - set(visited.keys()):
                size += len(list(visited)) + 1
            stack.pop()
            visited.popitem()

    return size

def count_total_paths(func_name, client, **kwargs):
    if client is None:
        return None

    graph = client.get_func_cfg(func_name)

    # count number of paths
    start_nodes = [node for node in graph.nodes if graph.in_degree(node) == 0]
    start_nodes = get_graph_root_nodes(graph) if len(start_nodes) > 1 else start_nodes
    start_node = start_nodes[0]
    end_nodes = {node for node in graph.nodes if graph.out_degree(node) == 0}
    count = _all_simple_paths_graph(graph, start_node, end_nodes, len(graph) - 1)

    return count



import os
import shutil
import sys
import time
from collections import defaultdict
from datetime import datetime
from pathlib import Path
import logging
from multiprocessing import Pool
from typing import List

import toml
from pyjoern import JoernClient, JoernServer, fast_cfgs_from_source
from pyjoern.mapping import cfg_root_node, correct_source_cfg_addrs

from sailreval import ALL_DECOMPILERS, ALL_METRICS, SAILR_DECOMPILERS, SAILR_METRICS, JOERNLESS_SERVER_METRICS
from sailreval.metrics import get_metric_function, POST_METRICS
from sailreval.metrics.ged_to_source import has_cfged_required_files, has_cfged_required_src_files
from sailreval.utils import bcolors, SAILR_DECOMPILATION_RESULTS_DIR, timeout, SAILR_MEASURE_RESULTS_DIR, WorkDirContext
from sailreval.decompilers.angr_dec import get_all_funcs
from sailreval.analysis.counting import decompilation_cache_to_results_format, collect_and_store_cache_data

from tqdm import tqdm

l = logging.getLogger(__name__)


def measure(filepath: Path, basename: str, metrics, functions, joern_port=9000, require_joern=True, client=None, **kwargs):
    if not metrics:
        return {}

    results = {}
    if not require_joern:
        client = None
    else:
        if client is None:
            client = JoernClient(filepath, port=joern_port, bin_name=basename)
    
    for metric in metrics:
        measure_func = get_metric_function(metric)
        results[metric] = measure_func(functions, client, **kwargs)

    return results

def _all_chached_or_joernless(basename: str, decompilers: list, metrics: list, cache_dir: Path):
    if decompilers is None or metrics is None:
        return False

    joernless_metrics = set(m for m in metrics if m in JOERNLESS_SERVER_METRICS)
    if len(joernless_metrics) == len(metrics):
        return True

    # some metrics is JOERNless, we must have a cache to be either not cached or joernless
    if cache_dir is None:
        l.debug("JOERN DIR IS NONE")
        return False

    cache_dir = Path(cache_dir)
    if not cache_dir.exists():
        l.debug("JOERN DIR does not exist!")
        return False

    cached_file = cache_dir / f"{basename}.toml"
    if not cached_file.exists():
        l.debug(f"JOERN cached filed does not exist: {cached_file}!")
        return False

    with open(cached_file, "r") as fp:
        cached_data = toml.load(fp)

    cached_metrics = set()
    for metric in metrics:
        for decompiler in decompilers:
            if metric not in cached_data.get(decompiler, {}):
                l.debug(f"Metric not in {decompiler} for {basename}, we must recompute metrics!")
                break
        else:
            cached_metrics.add(metric)

    l.debug(f"Cached metrics: {cached_metrics}")
    return len(cached_metrics.union(joernless_metrics)) == len(metrics)


def measure_files_with_joern_server(
    file_dir: Path, basename: str, decompilers=None, metrics=None, functions=None,
    cores=1, joern_port=9000, cache_dir=None, force_joern_skip=False, results_dir=None
):
    def _measure_files_with_joern_server():
        can_skip = _all_chached_or_joernless(basename, decompilers, metrics, cache_dir) or force_joern_skip
        if can_skip:
            l.debug(f"JOERN server not needed for {basename}")
            measure_files(file_dir, basename, decompilers=decompilers, metrics=metrics, functions=functions, cores=cores, joern_port=joern_port, require_joern=False, cache_dir=cache_dir)
        else:
            try:
                with JoernServer(port=joern_port):
                    measure_files(file_dir, basename, decompilers=decompilers, metrics=metrics, functions=functions, cores=cores, joern_port=joern_port, cache_dir=cache_dir)
            except Exception:
                return

    if results_dir is not None:
        with WorkDirContext(results_dir):
            _measure_files_with_joern_server()
    else:
        _measure_files_with_joern_server()


def _load_cache(cache_dir: Path, basename: str):
    if cache_dir is None:
        return {}

    cache_dir = Path(cache_dir).absolute()
    binary_cache = cache_dir / f"{basename}.toml"
    if not binary_cache.exists():
        return {}

    with open(binary_cache, "r") as fp:
        cached_results = toml.load(fp)

    return cached_results


def _cache_has_all_metrics(cache_results: dict, metrics: list, decompilers: list):
    for dec in decompilers:
        for metric in metrics:
            if metric not in cache_results.get(dec, {}):
                return False

    return True


def measure_files(file_dir: Path, basename: str, decompilers=None, metrics=None, functions=None, cores=1, joern_port=9000, single_metric_timeout=100, require_joern=True, cache_dir=None):
    decompilers = decompilers or ALL_DECOMPILERS
    metrics = metrics or ALL_METRICS
    #l.info(f"Starting evaluation on {basename} for metrics {metrics}")

    start_dt = datetime.now()
    file_dir = Path(file_dir).absolute()
    
    # get target files
    target_files = list(file_dir.rglob(f"*_{basename}.c"))
    target_binary = file_dir.joinpath(f"{basename}.o").absolute()
    all_source_types = [f"{dec}_{basename}.c" for dec in decompilers]
    for tfile in list(target_files):
        if tfile.name not in all_source_types:
            target_files.remove(tfile)

    # check the cache before doing expensive operations
    cached_results = _load_cache(cache_dir, basename)
    if _cache_has_all_metrics(cached_results, metrics, decompilers):
        l.info(f"All results are cached for {file_dir}/{basename}, skipping...")
        cached_results["binary"] = basename
        cached_results["total_time"] = 0
        cached_results["timeout"] = False
        with open(f"{basename}.toml", "w") as fp:
            toml.dump(cached_results, fp)

        return

    # generate default options for the set
    functions = functions or get_all_funcs(target_binary)
    if not functions:
        l.warning(f"Failed to find any functions for analysis and none were provided for {file_dir}/{basename}, using cache...")
        cache_dir = Path(cache_dir).absolute()
        binary_cache = cache_dir / f"{basename}.toml"
        if cache_dir is None or not binary_cache.exists():
            l.critical(f"Tried to use cache for functions on {file_dir}/{basename}, but none existed. Metrics will be skipped.")
            return

        with open(binary_cache, "r") as fp:
            cached_results = toml.load(fp)
        
        try:
            goto_metrics = cached_results[SAILR_DECOMPILERS.SOURCE_CODE][SAILR_METRICS.GOTO_COUNT]
        except KeyError:
            l.critical(f"Tried to use cache for functions on {file_dir}/{basename}, but no source with goto count existed")
            return

        functions = list(goto_metrics.keys()) 

    require_cfgs = any(
        m in metrics for m in (
            SAILR_METRICS.CFGED, SAILR_METRICS.GED_UPPERBOUND, SAILR_METRICS.GED_MAX, SAILR_METRICS.BLOCK_COUNT,
            SAILR_METRICS.CODE_COMPLEXITY, SAILR_METRICS.GED_EXACT
        )
    )

    # collect the source CFGs first if they are needed
    source_cfgs, dec_cfgs = {}, {}
    if require_cfgs and SAILR_DECOMPILERS.SOURCE_CODE in decompilers:
        for tfile in target_files:
            if tfile.name.startswith(SAILR_DECOMPILERS.SOURCE_CODE):
                if require_cfgs:
                    if has_cfged_required_src_files(tfile, target_binary):
                        linemaps_path = tfile.with_suffix(".linemaps")
                        extracted_cfgs = fast_cfgs_from_source(tfile.absolute())
                        if extracted_cfgs:
                            source_cfgs = correct_source_cfg_addrs(
                                extracted_cfgs,
                                linemaps_path,
                            )
                        else:
                            l.critical(f"JOERN failed to extract CFGS on {file_dir}/{basename}")
                            source_cfgs = {}
                    else:
                        l.critical(f"JOERN gave no source CFG for {file_dir}/{basename}, will skip all CFG metrics...")
                        source_cfgs = {}
                break

    results = defaultdict(dict)
    results["binary"] = basename
    had_timeout = False

    for target_file in target_files:
        filename = target_file.name
        dec_name = filename.split(f"_{basename}.c")[0]
        # skip targets that are not in requested decompiler set 
        if dec_name not in decompilers:
            continue

        # record which metric can be skipped
        cached_metric = []
        if dec_name in cached_results:
            cached_metric = list(cached_results[dec_name].keys())

        metrics_to_run = [m for m in metrics if m not in cached_metric]
        if metrics_to_run:
            # extract cfgs if needed
            if require_cfgs and has_cfged_required_files(target_file.absolute()):
                try:
                    dec_cfgs = fast_cfgs_from_source(target_file.absolute()) if dec_name != "source" else source_cfgs
                except Exception:
                    dec_cfgs = {}

            try:
                with timeout(seconds=single_metric_timeout*len(decompilers)):
                    results[dec_name] = measure(
                        target_file.absolute(), basename, metrics_to_run, functions, joern_port=joern_port,
                        dec_cfgs=dec_cfgs, source_cfgs=source_cfgs, binary_path=target_binary, decompiler=dec_name,
                        require_joern=require_joern
                    )
            except TimeoutError:
                l.warning(f"Measurement on {basename} for {decompilers} on {metrics} timedout! Will rerun once in next iter.")
                had_timeout = True

        # ready output
        if dec_name not in results:
            results[dec_name] = {}

        for metric in cached_metric:
            results[dec_name][metric] = cached_results[dec_name][metric]

    total_time = datetime.now().timestamp() - start_dt.timestamp()
    results["total_time"] = total_time
    results["timeout"] = had_timeout

    with open(f"{basename}.toml", "w") as fp:
        toml.dump(results, fp)

    #l.info(f"{bcolors.WARNING}{bcolors.BOLD}Completed evaluation on {basename} in {total_time} seconds.{bcolors.ENDC}.")


def _prep_measurement_directory(target_package_dir: Path, use_cache=True):
    """
    Verifies things exists, makes decompilation dir, may destroy old cache
    """
    sailr_dec_dir = target_package_dir.parent.absolute().joinpath(SAILR_DECOMPILATION_RESULTS_DIR).absolute()
    sailr_meas_dir = target_package_dir.absolute()
    if not sailr_meas_dir.exists():
        raise Exception(f"No sailr results dir for {target_package_dir}")

    all_binaries = list(sailr_dec_dir.glob("*.o"))
    binaries = all_binaries
    # if the user wants no cache
    if not use_cache and sailr_meas_dir is not None:
        # only do non_cached results
        for binary in all_binaries:
            basename = binary.with_suffix("").name
            if sailr_meas_dir.joinpath(f"{basename}.toml").exists():
                l.info(f"Found {basename} in measurements for {sailr_meas_dir} already, skipping...")
                continue

            binaries.append(binary)

    # destroy old cache if it exists (will recompute)
    old_cache = sailr_meas_dir / "cache"
    if old_cache.exists():
        shutil.rmtree(old_cache)
    res_cache_dir = old_cache
    os.mkdir(res_cache_dir)

    # recreate data in the cache based on prev results and the decompilers toml data
    res_cache_dir = collect_and_store_cache_data(sailr_meas_dir, sailr_dec_dir)

    # get the binary and the basename
    binaries_to_run = [] # (sailr_res_dir, sailr_dec_dir, basename, cache_dir)
    for binary_path in binaries:
        basename = binary_path.with_suffix("").name
        file_with_basename = list(sailr_dec_dir.rglob(f"*{basename}.c"))
        if not file_with_basename:
            l.warning(f"{bcolors.FAIL}{bcolors.BOLD} Binary {sailr_dec_dir}/{basename} either does not exist or has no source skipping... {bcolors.ENDC}")
            continue

        binaries_to_run.append((sailr_meas_dir, sailr_dec_dir, basename, res_cache_dir))

    return binaries_to_run


def batch_measure_targets(target_package_dirs: List[Path], decompilers=None, metrics=None, cores=1, first_port=9000, no_joern_server=False, use_cache=True):
    if metrics is not None:
        # remove all the post metrics since they are computed after eval
        metrics = [m for m in metrics if m not in POST_METRICS]

    l.info(f"Prepping batch measurement of {len(target_package_dirs)} directories...")
    jobs_no_port = [] # (sailr_res_dir, sailr_dec_dir, basename, cache_dir, joern_port)
    # first, extract all the binaries - source pairs
    for target_package_dir in target_package_dirs:
        # get jobs, which lack the joern port currently
        jobs_no_port += _prep_measurement_directory(target_package_dir, use_cache=use_cache)

    # now, we have all the jobs, but we needs port numbers
    jobs = [
        (sailr_res_dir, sailr_dec_dir, basename, cache_dir, first_port+i)
        for (i, (sailr_res_dir, sailr_dec_dir, basename, cache_dir)) in enumerate(jobs_no_port)
    ]

    # now, we can run the jobs
    l.info(f"Starting batch measurement of {len(jobs)} binaries across {len(target_package_dirs)} directories...")
    l.info(f"Evaluating with {cores} cores on {decompilers} with {metrics}, {'WITHOUT' if no_joern_server else 'with'} JOERN servers...")
    time.sleep(5)

    with Pool(processes=cores) as pool, tqdm(total=len(jobs), file=sys.stdout) as pbar:
        work = {}
        for job_args in jobs:
            sailr_res_dir, sailr_dec_dir, basename, cache_dir, joern_port = job_args
            w = pool.apply_async(
                measure_files_with_joern_server,
                args=(sailr_dec_dir, basename),
                kwds={"decompilers": decompilers, "metrics": metrics, "joern_port": joern_port, "cache_dir": cache_dir,
                      "force_joern_skip": no_joern_server, "results_dir": sailr_res_dir},
                callback=lambda _: pbar.update(1)
            )
            work[w] = job_args

        pool.close()
        for w, job_args in work.items():
            try:
                # every binary gets a max of 20 minutes
                w.get(timeout=60 * 20)
            except Exception as e:
                l.critical(f"Job {job_args} failed with exception {e}")
                pbar.update(1)
                continue

    l.info(f"Finished batch measurement of {len(jobs)} binaries across {len(target_package_dirs)} directories!")
